from http import HTTPStatus


def http_status_page(status: HTTPStatus, message: str) -> str:
    return """<!DOCTYPE html>
<html>
    <head>
        <title>{0.value} &mdash; {0.phrase}</title>
    </head>
    <body>
        <h1>{0.value} &mdash; {0.phrase}</h1>
        <p>{1}</p>
    </body>
</html>
""".format(status, message)


def created_at_page(url: str) -> str:
    message = 'Created at <a href="{0}">{0}</a>.'.format(url)
    return http_status_page(HTTPStatus.CREATED, message)


def see_other_page(url: str) -> str:
    message = 'Please see <a href="{0}">{0}</a>.'.format(url)
    return http_status_page(HTTPStatus.SEE_OTHER, message)



from typing import Callable, Tuple, List, Dict, Any, Iterable

from werkzeug.wrappers import Request

# (name, value)
HeaderType = Tuple[str, str]

EnvironmentType = Dict[str, Any]

# (body) -> None
StartResponseReturnType = Callable[[bytes], None]

# (status, headers) -> write
StartResponseType = Callable[[str, List[HeaderType]], StartResponseReturnType]

# (request, path, start_response) -> response
RouteHandler = \
    Callable[[Request, List[Any], StartResponseType], Iterable[bytes]]

# (method, path, callback)
RouteType = Tuple[str, str, RouteHandler]

# (request, previous_args, path_part) -> result
RouteTemplateHandler = Callable[[Request, List[Any], str], Any]



from enum import Enum
from http import HTTPStatus
import logging
import re
from typing import Iterable, List, Dict, Any, Tuple, Iterator, cast

from werkzeug.exceptions import NotFound, MethodNotAllowed, HTTPException
from werkzeug.wrappers import Request

from rouver.html import http_status_page
from rouver.response import respond_with_html
from rouver.types import StartResponseType, EnvironmentType, RouteType, \
    RouteHandler, RouteTemplateHandler


LOGGER_NAME = "rouver"


class _TemplatePartType(Enum):

    STATIC = 1
    PATTERN = 2


_TemplateHandlerDict = Dict[str, RouteTemplateHandler]
_RouteTemplatePart = Tuple[_TemplatePartType, str]


def _split_path(s: str) -> List[str]:
    if not s:
        return []
    return s.split("/")


class Router:

    def __init__(self) -> None:
        self._routes = []  # type: List[RouteType]
        self._template_handlers = {}  # type: _TemplateHandlerDict
        self.error_handling = True

    def __call__(self, environment: EnvironmentType,
                 start_response: StartResponseType) -> Iterable[bytes]:
        request = Request(environment)
        handlers = [
            _RouteHandler(r, self._template_handlers) for r in self._routes]
        try:
            return _dispatch(request, start_response, handlers,
                             self._template_handlers)
        except Exception:
            if self.error_handling:
                logging.getLogger(LOGGER_NAME).exception(
                    "error while handling request")
                return _respond_internal_server_error(start_response)
            else:
                raise

    def add_routes(self, routes: List[RouteType]) -> None:
        self._routes.extend(routes)

    def add_template_handler(self, name: str, handler: RouteTemplateHandler) \
            -> None:
        self._template_handlers[name] = handler


_pattern_re = re.compile(r"^{(.*)}$")


def _parse_route_template_part(
        part: str, template_handlers: _TemplateHandlerDict) \
        -> _RouteTemplatePart:
    m = _pattern_re.match(part)
    if m:
        name = m.group(1)
        if name not in template_handlers:
            msg = "unknown template path handler: '{}'".format(name)
            raise KeyError(msg) from None
        return _TemplatePartType.PATTERN, name
    else:
        return _TemplatePartType.STATIC, part


class _RouteHandler:

    def __init__(self, route: RouteType,
                 template_handlers: _TemplateHandlerDict) -> None:
        self.path = self._parse_path(route[0], template_handlers)
        self.method = route[1]
        self.handler = route[2]

    @staticmethod
    def _parse_path(path_string: str,
                    template_handlers: _TemplateHandlerDict) \
            -> List[_RouteTemplatePart]:
        parts = _split_path(path_string)
        return [_parse_route_template_part(part, template_handlers)
                for part in parts]


def _dispatch(request: Request, start_response: StartResponseType,
              handlers: List[_RouteHandler],
              template_handlers: _TemplateHandlerDict) -> Iterable[bytes]:

    def find_route_and_call_handler() -> Iterable[bytes]:
        try:
            handler, arguments = find_route()
        except NotFound:
            return _respond_not_found(request, start_response)
        except MethodNotAllowed as exc:
            return _respond_method_not_allowed(
                start_response, request.method, exc.valid_methods)
        else:
            return _call_handler(handler, arguments)

    def find_route() -> Tuple[RouteHandler, List[Any]]:
        arguments = _RouteArguments(request, template_handlers)
        matching_routes = [
            _RouteMatcher(h, request.path, arguments) for h in handlers]

        matching_paths = [m for m in matching_routes if m.matches]
        if not matching_paths:
            raise NotFound()

        matching_routes = \
            [m for m in matching_paths if m.method == request.method]
        if not matching_routes:
            valid_methods = sorted(m.method for m in matching_paths)
            raise MethodNotAllowed(valid_methods)

        assert len(matching_routes) == 1
        route = matching_routes[0]
        return route.handler, route.path_args

    def _call_handler(handler: RouteHandler, path_args: List[Any]) \
            -> Iterable[bytes]:
        try:
            return handler(request, path_args, start_response)
        except HTTPException as exc:
            return _respond_http_exception(start_response, exc)

    return find_route_and_call_handler()


class _RouteArguments:

    def __init__(self, request: Request,
                 template_handlers: _TemplateHandlerDict) -> None:
        self._request = request
        self._handlers = template_handlers
        self._cache = {}  # type: Dict[Tuple[str, str], Any]

    def parse_argument(self, paths: List[Any], name: str, path: str) -> Any:
        key = name, path
        if key not in self._cache:
            handler = self._handlers[name]
            self._cache[key] = handler(self._request, paths, path)
        return self._cache[key]


class _RouteMatcher:

    def __init__(self, handler: _RouteHandler, path: str,
                 arguments: _RouteArguments) -> None:
        self._handler = handler
        self.method = handler.method
        self.handler = handler.handler
        self._path = _split_path(path[1:])
        self._arguments = arguments
        self.path_args = []  # type: List[Any]
        if self._do_path_and_tmpl_differ_in_length():
            self.matches = False
        else:
            self.matches = self._match_parts()

    def _do_path_and_tmpl_differ_in_length(self):
        return len(self._handler.path) != len(self._path)

    def _match_parts(self) -> bool:
        for tmpl_part, path_part in self._path_compare_iter():
            tmpl_type, text = tmpl_part
            if tmpl_type == _TemplatePartType.STATIC:
                if text != path_part:
                    return False
            elif tmpl_type == _TemplatePartType.PATTERN:
                try:
                    arg = self._arguments.parse_argument(
                        self.path_args, text, path_part)
                except ValueError:
                    return False
                self.path_args.append(arg)
        return True

    def _path_compare_iter(self) -> Iterator[Tuple[_RouteTemplatePart, str]]:
        return zip(self._handler.path, self._path)


def _respond_not_found(request: Request, start_response: StartResponseType) \
        -> Iterable[bytes]:
    path = cast(str, request.environ.get("PATH_INFO", ""))
    message = "Path '{}' not found.".format(path)
    page = http_status_page(HTTPStatus.NOT_FOUND, message)
    return respond_with_html(
        start_response, page, status=HTTPStatus.NOT_FOUND)


def _respond_method_not_allowed(start_response: StartResponseType,
                                method: str, allowed_methods: List[str]) \
        -> Iterable[bytes]:
    method_string = " or ".join(allowed_methods)
    message = "Method '{}' not allowed. Please try {}.".format(
        method, method_string)
    html = http_status_page(HTTPStatus.METHOD_NOT_ALLOWED, message)
    return respond_with_html(
        start_response, html, status=HTTPStatus.METHOD_NOT_ALLOWED,
        extra_headers=[("Allow", ", ".join(allowed_methods))])


def _respond_internal_server_error(start_response: StartResponseType) \
        -> Iterable[bytes]:
    html = http_status_page(HTTPStatus.INTERNAL_SERVER_ERROR,
                            message="Internal server error.")
    return respond_with_html(start_response, html,
                             status=HTTPStatus.INTERNAL_SERVER_ERROR)


def _respond_http_exception(start_response: StartResponseType,
                            exception: HTTPException) -> Iterable[bytes]:
    status = HTTPStatus(exception.code)
    html = http_status_page(status, message=exception.get_description())
    return respond_with_html(start_response, html, status=status)



import collections
from http import HTTPStatus
from typing import List, Any, Union, Iterator

from werkzeug.wrappers import Request

from rouver.args import ArgumentTemplate, ArgumentDict, parse_args
from rouver.response import \
    respond, respond_with_json, respond_with_html, created_at, see_other
from rouver.types import StartResponseType, HeaderType


class RouteHandlerBase(collections.Iterable):

    """Base class for rouver route handlers.

    Sub-classes of RouteHandlerBase can act as route handlers. They provide
    convenient, opaque access to several rouver services.

    Implementations must implement the __iter__() method.

    >>> from rouver.router import Router
    >>> class MyRouteHandler(RouteHandlerBase):
    ...     def __iter__(self):
    ...         return self.respond_with_html("<div>Hello World!</div>")
    >>> class MyRouter(Router):
    ...     def __init__(self):
    ...         super().__init__()
    ...         self.add_routes([
    ...             ("my-route", "GET", MyRouteHandler),
    ...         ])
    """

    def __init__(self, request: Request, path_args: List[Any],
                 start_response: StartResponseType) -> None:
        self.request = request
        self.path_args = path_args
        self.start_response = start_response

    def __iter__(self) -> Iterator[bytes]:
        raise NotImplementedError()

    def parse_args(self, argument_template: List[ArgumentTemplate]) \
            -> ArgumentDict:
        return parse_args(self.request.environ, argument_template)

    def respond(self, extra_headers: List[HeaderType] = []) \
            -> Iterator[bytes]:
        return respond(self.start_response, extra_headers=extra_headers)

    def respond_with_json(
            self, json: Union[str, bytes, Any], *,
            status: HTTPStatus = HTTPStatus.OK,
            extra_headers: List[HeaderType] = []) -> Iterator[bytes]:
        return respond_with_json(
            self.start_response, json,
            status=status, extra_headers=extra_headers)

    def respond_with_html(
            self, html: str, *, status: HTTPStatus = HTTPStatus.OK,
            extra_headers: List[HeaderType] = []) \
            -> Iterator[bytes]:
        return respond_with_html(
            self.start_response, html,
            status=status, extra_headers=extra_headers)

    def created_at(self, url_part: str) -> Iterator[bytes]:
        return created_at(self.request, self.start_response, url_part)

    def see_other(self, url_part: str) -> Iterator[bytes]:
        return see_other(self.request, self.start_response, url_part)



from http import HTTPStatus
from json import dumps as dumps_json
from typing import List, Union, Any, Iterator

from werkzeug.wrappers import Request

from rouver.html import created_at_page, see_other_page
from rouver.status import status_line
from rouver.types import StartResponseType, HeaderType


def respond(start_response: StartResponseType, *,
            status: HTTPStatus = HTTPStatus.OK,
            extra_headers: List[HeaderType] = []) -> Iterator[bytes]:

    """Prepare an empty WSGI response.

    >>> def handler(start_response, request):
    ...     return respond(start_response, status=HTTPStatus.ACCEPTED)

    The return value can be ignored to return an arbitrary response body.

    >>> def handler(start_response, request):
    ...     respond(start_response,
    ...             extra_headers=[("Content-Type", "text/plain")])
    ...     return [b"My Response"]
    """

    sl = status_line(status)
    start_response(sl, extra_headers)
    return iter([])


def respond_with_json(start_response: StartResponseType,
                      json: Union[str, bytes, Any], *,
                      status: HTTPStatus = HTTPStatus.OK,
                      extra_headers: List[HeaderType] = []) \
        -> Iterator[bytes]:

    """Prepare a JSON WSGI response.

    >>> def handler(start_response, request):
    ...     return respond_with_json(start_response, {"foo": "bar"})

    The JSON text to return can be supplied as a string, as an UTF-8-encoded
    bytestring, or as any object that can be serialized using json.dumps().

    The default response status of "200 OK" can be overridden with the
    "status" keyword argument.
    """

    sl = status_line(status)
    headers = \
        [("Content-Type", "application/json; charset=utf-8")] + extra_headers
    start_response(sl, headers)
    if isinstance(json, bytes):
        encoded = json
    elif isinstance(json, str):
        encoded = json.encode("utf-8")
    else:
        encoded = dumps_json(json).encode("utf-8")
    return iter([encoded])


def respond_with_html(start_response: StartResponseType, html: str, *,
                      status: HTTPStatus = HTTPStatus.OK,
                      extra_headers: List[HeaderType] = []) \
        -> Iterator[bytes]:

    """Prepare an HTML WSGI response.

    >>> def handler(start_response, request):
    ...     return respond_with_html(start_response, "<div>foo</div>")

    The default response status of "200 OK" can be overridden with the
    "status" keyword argument.
    """

    sl = status_line(status)
    headers = [("Content-Type", "text/html; charset=utf-8")] + extra_headers
    start_response(sl, headers)
    return iter([html.encode("utf-8")])


def created_at(request: Request, start_response: StartResponseType,
               url_part: str) -> Iterator[bytes]:

    """Prepare a 201 Created WSGI response with a Location header."""

    if url_part.startswith("/"):
        url_part = url_part[1:]
    url = request.host_url + url_part
    html = created_at_page(url)
    return respond_with_html(start_response, html, status=HTTPStatus.CREATED,
                             extra_headers=[("Location", url)])


def see_other(request: Request, start_response: StartResponseType,
              url_part: str) -> Iterator[bytes]:
    if url_part.startswith("/"):
        url_part = url_part[1:]
    url = request.host_url + url_part
    html = see_other_page(url)
    return respond_with_html(start_response, html, status=HTTPStatus.SEE_OTHER,
                             extra_headers=[("Location", url)])



from http import HTTPStatus
import logging
from typing import Iterable, List, Any
from unittest import TestCase

from asserts import assert_equal, assert_raises, assert_is_instance

from werkzeug.exceptions import Conflict
from werkzeug.wrappers import Request

from rouver.router import Router, LOGGER_NAME
from rouver.types import StartResponseType

from rouver_test.util import StartResponse, default_environment


def handle_success(_, __, start_response: StartResponseType) \
        -> Iterable[bytes]:
    start_response("200 OK", [])
    return [b""]


def handle_empty_path(_, path: List[str], start_response: StartResponseType) \
        -> Iterable[bytes]:
    assert_equal([], path)
    start_response("200 OK", [])
    return [b""]


def fail_if_called(_, __, ___) -> Iterable[bytes]:
    raise AssertionError("handler should not be called")


class RouterTest(TestCase):

    def setUp(self) -> None:
        self.router = Router()
        self.router.error_handling = False
        self.start_response = StartResponse()
        self.environment = default_environment()
        self.disable_logger()

    def disable_logger(self) -> None:
        logging.getLogger(LOGGER_NAME).disabled = True

    def handle_wsgi(self, method: str = "GET", path: str = "/") \
            -> Iterable[bytes]:
        self.environment["REQUEST_METHOD"] = method
        self.environment["PATH_INFO"] = path
        return self.router(self.environment, self.start_response)

    def test_no_routes(self) -> None:
        response = self.handle_wsgi("GET", "/foo/bar")
        self.start_response.assert_status(HTTPStatus.NOT_FOUND)
        self.start_response.assert_header_equals(
            "Content-Type", "text/html; charset=utf-8")
        html = b"".join(response).decode("utf-8")
        assert_equal("""<!DOCTYPE html>
<html>
    <head>
        <title>404 &mdash; Not Found</title>
    </head>
    <body>
        <h1>404 &mdash; Not Found</h1>
        <p>Path '/foo/bar' not found.</p>
    </body>
</html>
""", html)

    def test_handler_request(self) -> None:
        def handle(request: Request, _, start_response):
            assert_equal("test.example.com", request.host)
            start_response("200 OK", [])
            return [b""]

        self.router.add_routes([
            ("", "GET", handle),
        ])
        self.environment["SERVER_NAME"] = "test.example.com"
        self.handle_wsgi("GET", "")
        self.start_response.assert_status(HTTPStatus.OK)

    def test_empty_route(self) -> None:
        self.router.add_routes([
            ("", "GET", handle_empty_path),
        ])
        self.handle_wsgi("GET", "")
        self.start_response.assert_status(HTTPStatus.OK)

    def test_root_route(self) -> None:
        self.router.add_routes([
            ("", "GET", handle_empty_path),
        ])
        self.handle_wsgi("GET", "/")
        self.start_response.assert_status(HTTPStatus.OK)

    def test_first_level(self) -> None:
        self.router.add_routes([
            ("foo", "GET", handle_empty_path),
        ])
        self.handle_wsgi("GET", "/foo")
        self.start_response.assert_status(HTTPStatus.OK)

    def test_first_level_wrong_path(self) -> None:
        self.router.add_routes([
            ("foo", "GET", handle_empty_path),
        ])
        self.handle_wsgi("GET", "/bar")
        self.start_response.assert_status(HTTPStatus.NOT_FOUND)

    def test_level_mismatch_1(self) -> None:
        self.router.add_routes([
            ("foo/bar", "GET", handle_empty_path),
        ])
        self.handle_wsgi("GET", "/foo")
        self.start_response.assert_status(HTTPStatus.NOT_FOUND)

    def test_level_mismatch_2(self) -> None:
        self.router.add_routes([
            ("foo", "GET", handle_empty_path),
        ])
        self.handle_wsgi("GET", "/foo/bar")
        self.start_response.assert_status(HTTPStatus.NOT_FOUND)

    # Method Handling

    def test_wrong_method(self) -> None:
        self.router.add_routes([
            ("foo", "GET", fail_if_called),
            ("foo", "PUT", fail_if_called),
        ])
        self.handle_wsgi("POST", "/foo")
        self.start_response.assert_status(HTTPStatus.METHOD_NOT_ALLOWED)
        self.start_response.assert_header_equals("Allow", "GET, PUT")

    def test_call_right_method(self) -> None:
        self.router.add_routes([
            ("foo", "GET", fail_if_called),
            ("foo", "POST", handle_success),
            ("foo", "PUT", fail_if_called),
        ])
        self.handle_wsgi("POST", "/foo")
        self.start_response.assert_status(HTTPStatus.OK)

    # Path Templates

    def test_unknown_template(self) -> None:
        self.router.add_routes([
            ("foo/{unknown}/bar", "GET", fail_if_called),
        ])
        with assert_raises(KeyError):
            self.handle_wsgi("GET", "/foo/xyz/bar")

    def test_template(self) -> None:
        def handle(_, path, start_response):
            assert_equal(["xyzxyz"], path)
            start_response("200 OK", [])
            return [b""]

        def handle_path(request: Request, paths: List[Any], path: str) -> str:
            assert_is_instance(request, Request)
            assert_equal([], paths)
            return path * 2
        self.router.add_template_handler("handler", handle_path)

        self.router.add_routes([
            ("foo/{handler}/bar", "GET", handle),
        ])
        self.handle_wsgi("GET", "/foo/xyz/bar")
        self.start_response.assert_status(HTTPStatus.OK)

    def test_multiple_templates(self) -> None:
        def handle(_, path, start_response):
            assert_equal(["xyz", 123], path)
            start_response("200 OK", [])
            return [b""]

        def handle_path(_, paths: List[Any], __) -> int:
            assert_equal(["xyz"], paths)
            return 123
        self.router.add_template_handler("handler1", lambda _, __, ___: "xyz")
        self.router.add_template_handler("handler2", handle_path)

        self.router.add_routes([
            ("foo/{handler1}/bar/{handler2}", "GET", handle),
        ])
        self.handle_wsgi("GET", "/foo/xyz/bar/abc")
        self.start_response.assert_status(HTTPStatus.OK)

    def test_template_value_error(self) -> None:
        def raise_value_error(_, __, ___):
            raise ValueError()
        self.router.add_template_handler("handler", raise_value_error)

        self.router.add_routes([
            ("foo/{handler}/bar", "GET", fail_if_called),
        ])
        self.handle_wsgi("GET", "/foo/xyz/bar")
        self.start_response.assert_status(HTTPStatus.NOT_FOUND)

    def test_template_call_once_per_value(self) -> None:
        calls = 0

        def increase_count(_, __, ___):
            nonlocal calls
            calls += 1
        self.router.add_template_handler("handler", increase_count)

        self.router.add_routes([
            ("foo/{handler}/bar", "GET", fail_if_called),
            ("foo/{handler}/baz", "GET", handle_success),
        ])
        self.handle_wsgi("GET", "/foo/xyz/baz")
        assert_equal(1, calls)

    def test_template_call_twice_for_differing_values(self) -> None:
        calls = 0

        def increase_count(_, __, ___):
            nonlocal calls
            calls += 1
        self.router.add_template_handler("handler", increase_count)

        self.router.add_routes([
            ("foo/{handler}/bar", "GET", fail_if_called),
            ("foo/xyz/{handler}", "GET", handle_success),
        ])
        self.handle_wsgi("GET", "/foo/xyz/baz")
        assert_equal(2, calls)

    # Error Handling

    def test_template_key_error_with_error_handling(self) -> None:
        def raise_key_error(_, __, ___):
            raise KeyError()
        self.router.add_template_handler("handler", raise_key_error)

        self.router.add_routes([
            ("foo/{handler}/bar", "GET", fail_if_called),
        ])
        self.router.error_handling = True
        self.handle_wsgi("GET", "/foo/xyz/bar")
        self.start_response.assert_status(HTTPStatus.INTERNAL_SERVER_ERROR)

    def test_template_key_error_without_error_handling(self) -> None:
        def raise_key_error(_, __, ___):
            raise KeyError()
        self.router.add_template_handler("handler", raise_key_error)

        self.router.add_routes([
            ("foo/{handler}/bar", "GET", fail_if_called),
        ])
        self.router.error_handling = False
        with assert_raises(KeyError):
            self.handle_wsgi("GET", "/foo/xyz/bar")

    def test_handler_key_error_with_error_handling(self) -> None:
        def handle(_, __, ___):
            raise KeyError()

        self.router.error_handling = True
        self.router.add_routes([
            ("foo", "GET", handle),
        ])
        self.handle_wsgi("GET", "/foo")
        self.start_response.assert_status(HTTPStatus.INTERNAL_SERVER_ERROR)

    def test_handler_key_error_without_error_handling(self) -> None:
        def handle(_, __, ___):
            raise KeyError()

        self.router.add_routes([
            ("foo", "GET", handle),
        ])
        self.router.error_handling = False
        with assert_raises(KeyError):
            self.handle_wsgi("GET", "/foo")

    def test_http_error(self) -> None:
        def handle(_, __, ___):
            raise Conflict()

        self.router.error_handling = False
        self.router.add_routes([
            ("foo", "GET", handle),
        ])
        response = self.handle_wsgi("GET", "/foo")
        self.start_response.assert_status(HTTPStatus.CONFLICT)
        html = b"".join(response).decode("utf-8")
        assert html.startswith("<!DOCTYPE html>")



from http import HTTPStatus

from unittest import TestCase

from asserts import assert_equal

from rouver.status import status_line


class StatusLineTest(TestCase):

    def test_status_line(self) -> None:
        sl = status_line(HTTPStatus.NO_CONTENT)
        assert_equal("204 No Content", sl)



import cgi
from enum import Enum
from io import BytesIO
from typing import Callable, Any, Tuple, Dict, List, Union, IO

from rouver.exceptions import ArgumentsError
from rouver.types import EnvironmentType


class Multiplicity(Enum):

    ANY = "0+"
    REQUIRED_ANY = "1+"
    REQUIRED = "1"
    OPTIONAL = "0-1"


ArgumentValueParser = Callable[[str], Any]
ArgumentValueType = Union[ArgumentValueParser, str]
ArgumentTemplate = Tuple[str, ArgumentValueType, Multiplicity]
ArgumentDict = Dict[str, Any]


class _ArgumentError(Exception):

    pass


class FileArgument:

    """File argument result.

    This is a file-like object, containing a bytes. It has additional
    fields "filename" and "content_type", as supplied by the client.

    >>> args = parse_args(..., [
    ...     ("file-arg", "file", Multiplicity.REQUIRED),
    ... ])
    >>> args["file-arg"].filename
    'my-file.txt'
    >>> args["file-arg"].content_type
    'text/plain'
    >>> args["file-arg"].read()
    b'file content'
    """

    def __init__(self, stream: IO[bytes], filename: str, content_type: str) \
            -> None:
        self._stream = stream
        self.filename = filename
        self.content_type = content_type

    def __getattr__(self, name: str) -> Any:
        return getattr(self._stream, name)


class CGIFileArgument(FileArgument):

    def __init__(self, value: cgi.FieldStorage) -> None:
        content_type = value.headers.get(
            "content-type", "application/octet-stream").split(";")[0]
        super().__init__(value.file, value.filename, content_type)
        # We keep a reference to the FieldStorage around, otherwise the
        # file will get closed in FieldStorage.__del__().
        self._value = value


def parse_args(environment: EnvironmentType,
               argument_template: List[ArgumentTemplate]) -> ArgumentDict:

    """Parse CGI/WSGI arguments and return an argument dict.

    argument_template is a list of (argname, value_parser, multiplicity)
    tuples. argname is the argument name that is expected to be provided.

    If multiplicity is REQUIRED and the argument was not supplied, an
    ArgumentsError is raised. If multiplicity is OPTIONAL and the
    argument was not supplied, the returned dict does not contain a key
    for this argument. multiplicity can also be ANY, in which case
    the key may be supplied any number of times, or REQUIRED_ANY, which works
    like ANY, but requires the argument to be provided at least once.

    value_parser is a callable that takes a single string argument. The value
    returned by this function will be used as the value of the returned
    dict. If the supplied string or list can't be parsed, this function
    should raise a ValueError exception. In this case, parse_args raises
    an ArgumentsError. If value_func is the string "file", the value of the
    argument is a FileArgument instance.

    The returned dict contains the names of all parsed arguments that were
    present in the CGI arguments as key. The value is the corresponding
    parsed argument value. In the case of ANY and REQUIRED_ANY arguments,
    the value is a list of parsed arguments values. Each ANY argument has an
    entry in the dict, even if the argument was not supplied. In this case the
    value is the empty list.

    If there is any error while parsing the arguments, an ArgumentsError
    will be raised.

    >>> from io import StringIO
    >>> environment = {
    ...     "QUERY_STRING": "key=value&multi=1&multi=2",
    ...     "wsgi.input": StringIO(""),
    ... }
    >>> args = parse_args(environment, [
    ...     ("key", str, Multiplicity.REQUIRED),
    ...     ("multi", int, Multiplicity.ANY),
    ...     ("optional", str, Multiplicity.OPTIONAL),
    ... ])
    >>> args["key"]
    'value'
    >>> args["multi"]
    [1, 2]
    >>> "optional" in args
    False
    >>> parse_args(environment, [("key", int, Multiplicity.REQUIRED)])
    Traceback (most recent call last):
    ...
    ArgumentsError: 400 Bad Request
    >>> parse_args(environment, [("missing", str, Multiplicity.REQUIRED)])
    Traceback (most recent call last):
    ...
    ArgumentsError: 400 Bad Request
    >>>

    """

    arguments = cgi.FieldStorage(
        environment["wsgi.input"], environ=environment,
        keep_blank_values=True)

    errors = {}  # type: Dict[str, str]
    parsed_arguments = {}  # type: ArgumentDict

    def parse_template(name: str, value_parser: ArgumentValueType,
                       multiplicity: Multiplicity) -> None:
        cls = _VALUE_PARSER_CLASSES[multiplicity]
        argument_parser = cls(arguments, name, value_parser)
        if argument_parser.should_parse():
            try:
                parsed_arguments[name] = argument_parser.parse()
            except _ArgumentError as exc:
                errors[name] = exc.args[0]

    for tmpl in argument_template:
        parse_template(*tmpl)

    if len(errors) > 0:
        raise ArgumentsError(errors)
    return parsed_arguments


class _ValueParserWrapper:

    def parse_from_string(self, s: str) -> Any:
        raise NotImplementedError()

    def parse_from_file(self, value: cgi.FieldStorage) -> Any:
        raise NotImplementedError()


class _FunctionValueParser(_ValueParserWrapper):

    def __init__(self, value_parser: ArgumentValueParser) -> None:
        super().__init__()
        self.value_parser = value_parser

    def parse_from_string(self, value: str) -> Any:
        try:
            return self.value_parser(value)
        except ValueError as exc:
            raise _ArgumentError(str(exc)) from exc

    def parse_from_file(self, value: cgi.FieldStorage) -> Any:
        content = value.file.read()
        decoded = \
            content.decode("utf-8") if isinstance(content, bytes) else content
        return self.parse_from_string(decoded)


class _FileArgumentValueParser(_ValueParserWrapper):

    def parse_from_string(self, value: str) -> FileArgument:
        stream = BytesIO(value.encode("utf-8"))
        return FileArgument(stream, "", "application/octet-stream")

    def parse_from_file(self, value: cgi.FieldStorage) -> CGIFileArgument:
        return CGIFileArgument(value)


def _create_argument_value_parser(value_parser: ArgumentValueType) \
        -> _ValueParserWrapper:
    if value_parser == "file":
        return _FileArgumentValueParser()
    elif callable(value_parser):
        return _FunctionValueParser(value_parser)
    else:
        raise TypeError("invalid value parser '{}'".format(value_parser))


class _ArgumentParser:

    def __init__(self, args: cgi.FieldStorage, name: str,
                 value_parser: ArgumentValueType) -> None:
        self.args = args
        self.name = name
        self.value_parser = _create_argument_value_parser(value_parser)

    def should_parse(self) -> bool:
        return True

    def parse(self) -> Any:
        raise NotImplementedError()


class _SingleArgumentParser(_ArgumentParser):

    def parse(self) -> Any:
        raise NotImplementedError()

    def parse_single_arg(self) -> Any:
        if self.is_argument_a_file:
            file = self.args[self.name]
            return self.value_parser.parse_from_file(file)
        else:
            value = self.args.getfirst(self.name)
            return self.value_parser.parse_from_string(value)

    @property
    def is_argument_a_file(self) -> bool:
        argv = self.args[self.name]
        return hasattr(argv, "file") and argv.file is not None


class _RequiredArgumentParser(_SingleArgumentParser):

    def parse(self) -> Any:
        if self.name not in self.args:
            raise _ArgumentError("mandatory argument missing")
        return self.parse_single_arg()


class _OptionalArgumentParser(_SingleArgumentParser):

    def should_parse(self) -> bool:
        if self.name not in self.args:
            return False
        return True

    def parse(self) -> Any:
        assert self.should_parse()
        return self.parse_single_arg()


class _MultiArgumentParser(_ArgumentParser):

    def parse(self) -> List[Any]:
        values = self.args.getlist(self.name)
        return [self.value_parser.parse_from_string(v) for v in values]


class _AtLeastOneArgumentParser(_MultiArgumentParser):

    def parse(self) -> List[Any]:
        values = super().parse()
        if not values:
            raise _ArgumentError("mandatory argument missing")
        return values


_VALUE_PARSER_CLASSES = {
    Multiplicity.ANY: _MultiArgumentParser,
    Multiplicity.REQUIRED_ANY: _AtLeastOneArgumentParser,
    Multiplicity.REQUIRED: _RequiredArgumentParser,
    Multiplicity.OPTIONAL: _OptionalArgumentParser,
}



from http import HTTPStatus
from unittest import TestCase

from asserts import assert_equal, assert_is_instance
from collections import Iterator

from werkzeug.wrappers import Request

from rouver.response import \
    respond, respond_with_json, respond_with_html, created_at, see_other

from rouver_test.util import StartResponse, default_environment


class RespondTest(TestCase):

    def test_default_status(self) -> None:
        sr = StartResponse()
        respond(sr)
        sr.assert_status(HTTPStatus.OK)

    def test_custom_status(self) -> None:
        sr = StartResponse()
        respond(sr, status=HTTPStatus.NOT_ACCEPTABLE)
        sr.assert_status(HTTPStatus.NOT_ACCEPTABLE)

    def test_extra_headers(self) -> None:
        sr = StartResponse()
        respond(sr, extra_headers=[
            ("X-Custom-Header", "Foobar"),
        ])
        sr.assert_header_equals("X-Custom-Header", "Foobar")

    def test_response(self) -> None:
        sr = StartResponse()
        response = respond(sr)
        assert_equal(b'', b"".join(response))

    def test_return_value_is_iterator(self) -> None:
        sr = StartResponse()
        response = respond(sr)
        assert_is_instance(response, Iterator)


class RespondWithJSONTest(TestCase):

    def test_default_status(self) -> None:
        sr = StartResponse()
        respond_with_json(sr, {})
        sr.assert_status(HTTPStatus.OK)

    def test_custom_status(self) -> None:
        sr = StartResponse()
        respond_with_json(sr, {}, status=HTTPStatus.NOT_ACCEPTABLE)
        sr.assert_status(HTTPStatus.NOT_ACCEPTABLE)

    def test_content_type(self) -> None:
        sr = StartResponse()
        respond_with_json(sr, {})
        sr.assert_header_equals(
            "Content-Type", "application/json; charset=utf-8")

    def test_extra_headers(self) -> None:
        sr = StartResponse()
        respond_with_json(sr, {}, extra_headers=[
            ("X-Custom-Header", "Foobar"),
        ])
        sr.assert_header_equals("X-Custom-Header", "Foobar")

    def test_json_as_bytes(self) -> None:
        sr = StartResponse()
        response = respond_with_json(sr, b'{"foo": 3}')
        assert_equal(b'{"foo": 3}', b"".join(response))

    def test_json_as_str(self) -> None:
        sr = StartResponse()
        response = respond_with_json(sr, '{"föo": 3}')
        assert_equal('{"föo": 3}'.encode("utf-8"), b"".join(response))

    def test_json_as_object(self) -> None:
        sr = StartResponse()
        response = respond_with_json(sr, {"föo": 3})
        assert_equal(b'{"f\u00f6o": 3}', b"".join(response))

    def test_return_value_is_iterator(self) -> None:
        sr = StartResponse()
        response = respond_with_json(sr, {})
        assert_is_instance(response, Iterator)


class RespondWithHTMLTest(TestCase):

    def test_default_status(self) -> None:
        sr = StartResponse()
        respond_with_html(sr, "<div>Test</div>")
        sr.assert_status(HTTPStatus.OK)

    def test_custom_status(self) -> None:
        sr = StartResponse()
        respond_with_html(
            sr, "<div>Test</div>", status=HTTPStatus.NOT_ACCEPTABLE)
        sr.assert_status(HTTPStatus.NOT_ACCEPTABLE)

    def test_content_type(self) -> None:
        sr = StartResponse()
        respond_with_html(sr, "<div>Test</div>")
        sr.assert_header_equals("Content-Type", "text/html; charset=utf-8")

    def test_extra_headers(self) -> None:
        sr = StartResponse()
        respond_with_html(sr, "<div>Täst</div>", extra_headers=[
            ("X-Custom-Header", "Foobar"),
        ])
        sr.assert_header_equals("X-Custom-Header", "Foobar")

    def test_return_value(self) -> None:
        sr = StartResponse()
        response = respond_with_html(sr, "<div>Test</div>")
        assert_equal(b"<div>Test</div>", b"".join(response))

    def test_return_value_is_iterator(self) -> None:
        sr = StartResponse()
        response = respond_with_html(sr, "<div>Test</div>")
        assert_is_instance(response, Iterator)

    def test_return_value_encoding(self) -> None:
        sr = StartResponse()
        response = respond_with_html(sr, "<div>Täst</div>")
        assert_equal("<div>Täst</div>".encode("utf-8"), b"".join(response))


class CreatedAtTest(TestCase):

    def setUp(self) -> None:
        self.environment = default_environment()
        self.start_request = StartResponse()

    def test_headers(self) -> None:
        self.environment["SERVER_NAME"] = "www.example.com"
        request = Request(self.environment)
        created_at(request, self.start_request, "/foo/bar")
        self.start_request.assert_status(HTTPStatus.CREATED)
        self.start_request.assert_header_equals(
            "Content-Type", "text/html; charset=utf-8")
        self.start_request.assert_header_equals(
            "Location", "http://www.example.com/foo/bar")

    def test_url_without_leading_slash(self) -> None:
        self.environment["SERVER_NAME"] = "www.example.com"
        request = Request(self.environment)
        created_at(request, self.start_request, "foo/bar")
        self.start_request.assert_header_equals(
            "Location", "http://www.example.com/foo/bar")

    def test_html(self) -> None:
        request = Request(self.environment)
        response = created_at(request, self.start_request, "foo/bar")
        html = b"".join(response).decode("utf-8")
        assert html.startswith("<!DOCTYPE html>")

    def test_return_value_is_iterator(self) -> None:
        request = Request(self.environment)
        response = created_at(request, self.start_request, "foo/bar")
        assert_is_instance(response, Iterator)


class SeeOtherTest(TestCase):

    def setUp(self) -> None:
        self.environment = default_environment()
        self.start_request = StartResponse()

    def test_headers(self) -> None:
        self.environment["SERVER_NAME"] = "www.example.com"
        request = Request(self.environment)
        see_other(request, self.start_request, "/foo/bar")
        self.start_request.assert_status(HTTPStatus.SEE_OTHER)
        self.start_request.assert_header_equals(
            "Content-Type", "text/html; charset=utf-8")
        self.start_request.assert_header_equals(
            "Location", "http://www.example.com/foo/bar")

    def test_url_without_leading_slash(self) -> None:
        self.environment["SERVER_NAME"] = "www.example.com"
        request = Request(self.environment)
        see_other(request, self.start_request, "foo/bar")
        self.start_request.assert_header_equals(
            "Location", "http://www.example.com/foo/bar")

    def test_html(self) -> None:
        request = Request(self.environment)
        response = see_other(request, self.start_request, "foo/bar")
        html = b"".join(response).decode("utf-8")
        assert html.startswith("<!DOCTYPE html>")

    def test_return_value_is_iterator(self) -> None:
        request = Request(self.environment)
        response = see_other(request, self.start_request, "foo/bar")
        assert_is_instance(response, Iterator)



from http import HTTPStatus
from unittest import TestCase

from asserts import assert_equal

from rouver.html import http_status_page


class HTTPStatusPageTest(TestCase):

    def test_http_status_page(self) -> None:
        html = http_status_page(HTTPStatus.NOT_ACCEPTABLE, "Test message.")
        assert_equal("""<!DOCTYPE html>
<html>
    <head>
        <title>406 &mdash; Not Acceptable</title>
    </head>
    <body>
        <h1>406 &mdash; Not Acceptable</h1>
        <p>Test message.</p>
    </body>
</html>
""", html)



from http import HTTPStatus
from io import BytesIO, StringIO
import re
from typing import List

from asserts import assert_false, assert_regex, assert_equal, assert_true

from rouver.types import StartResponseReturnType, HeaderType, EnvironmentType

_status_re = re.compile(r"^(\d\d\d) (.*)$")


class StartResponse:

    def __init__(self) -> None:
        self.was_called = False
        self.status = ""
        self.headers = []  # type: List[HeaderType]

    def __call__(self, status: str, headers: List[HeaderType]) \
            -> StartResponseReturnType:
        assert_false(self.was_called, "start_response() called twice")
        assert_regex(status, _status_re)
        self.was_called = True
        self.status = status
        self.headers = headers
        return lambda s: None

    @property
    def status_code(self) -> int:
        self.assert_was_called()
        assert len(self.status) >= 3
        return int(self.status[:3])

    def assert_was_called(self) -> None:
        assert_true(self.was_called, "start_response() was not called")

    def assert_status(self, status: HTTPStatus) -> None:
        assert_equal(status.value, self.status_code)

    def assert_header_equals(self, name: str, value: str) -> None:
        header_value = self._find_header(name)
        assert_equal(value, header_value, "'{}': '{}' != '{}".format(
            name, value, header_value))

    def _find_header(self, name: str) -> str:
        self.assert_was_called()
        for (header_name, header_value) in self.headers:
            if header_name.lower() == name.lower():
                return header_value
        raise AssertionError("missing header '{}'".format(name))


def default_environment() -> EnvironmentType:
    return {
        "REQUEST_METHOD": "GET",
        "SERVER_NAME": "www.example.com",
        "SERVER_PORT": "80",
        "SERVER_PROTOCOL": "HTTP/1.1",
        "wsgi.version": (1, 0),
        "wsgi.url_scheme": "http",
        "wsgi.input": BytesIO(b""),
        "wsgi.errors": StringIO(),
        "wsgi.multithread": False,
        "wsgi.multiprocess": False,
        "wsgi.run_once": False,
    }



from typing import Dict

from werkzeug.exceptions import BadRequest


class ArgumentsError(BadRequest):

    """One or more CGI or JSON arguments were invalid.

    Requires a dictionary as argument, where keys are the argument names or
    JSON paths that have errors, and the values are messages describing the
    problem.

    >>> error = ArgumentsError({
    ...     "foo": "argument missing",
    ...     "bar": "int required, got 'baz'",
    ... })
    >>> len(error.arguments)
    2

    """

    def __init__(self, arguments: Dict[str, str]) -> None:
        super().__init__("invalid arguments")
        self.arguments = arguments



from unittest import TestCase

from asserts import assert_equal

from rouver.exceptions import ArgumentsError


class ArgumentsErrorTest(TestCase):

    def test_message(self) -> None:
        error = ArgumentsError({"foo": "bar"})
        assert_equal("400 Bad Request: invalid arguments", str(error))

    def test_arguments(self) -> None:
        error = ArgumentsError({"foo": "bar"})
        assert_equal({"foo": "bar"}, error.arguments)



from unittest import TestCase
from urllib.parse import quote_plus

from asserts import assert_equal, fail, assert_raises, assert_in
from io import BytesIO

from rouver.args import parse_args, Multiplicity
from rouver.exceptions import ArgumentsError

from rouver_test.util import default_environment


MULTIPART_BODY_TMPL = """--1234567890
Content-Disposition: form-data; name="{}"

{}
--1234567890--
"""

MULTIPART_FILE_BODY_TMPL = """--1234567890
Content-Disposition: form-data; name="{name}"; filename={filename}
Content-Type: {type}

{content}
--1234567890--
"""


class ParseArgsTest(TestCase):

    def setUp(self) -> None:
        self.env = default_environment()

    def add_path_argument(self, name: str, value: str) -> None:
        if "QUERY_STRING" not in self.env:
            self.env["QUERY_STRING"] = ""
        if self.env["QUERY_STRING"]:
            self.env["QUERY_STRING"] += "&"
        self.env["QUERY_STRING"] += \
            "{}={}".format(quote_plus(name), quote_plus(value))

    def setup_urlencoded_request(self, name: str, value: str) -> None:
        body = "{}={}".format(name, value).encode("utf-8")
        self.env["CONTENT_TYPE"] = \
            "application/x-www-form-urlencoded; charset=utf-8"
        self.env["CONTENT_LENGTH"] = str(len(body))
        self.env["wsgi.input"] = BytesIO(body)

    def setup_multipart_request(self, name: str, value: str) -> None:
        self.env["CONTENT_TYPE"] = \
            "multipart/form-data; boundary=1234567890"
        body = MULTIPART_BODY_TMPL.format(name, value).encode("utf-8")
        self.env["CONTENT_LENGTH"] = str(len(body))
        self.env["wsgi.input"] = BytesIO(body)

    def setup_multipart_file_request(
            self, name: str, filename: str, file_content: str,
            content_type) -> None:
        self.env["CONTENT_TYPE"] = \
            "multipart/form-data; boundary=1234567890"
        body = MULTIPART_FILE_BODY_TMPL.format(
            name=name, filename=filename, content=file_content,
            type=content_type).encode("utf-8")
        self.env["CONTENT_LENGTH"] = str(len(body))
        self.env["wsgi.input"] = BytesIO(body)

    def test_parse_nothing(self) -> None:
        args = parse_args(self.env, [])
        assert_equal({}, args)

    def test_invalid_value_parser(self) -> None:
        with assert_raises(TypeError):
            parse_args(self.env, [
                ("foo", "INVALID", Multiplicity.OPTIONAL),
            ])

    def test_parse_str_arg(self) -> None:
        self.add_path_argument("foo", "bar")
        args = parse_args(self.env, [
            ("foo", str, Multiplicity.REQUIRED),
        ])
        assert_equal({"foo": "bar"}, args)

    def test_parse_unicode_arg(self) -> None:
        self.add_path_argument("föo", "bär")
        args = parse_args(self.env, [
            ("föo", str, Multiplicity.REQUIRED),
        ])
        assert_equal({"föo": "bär"}, args)

    def test_parse_int_arg(self) -> None:
        self.add_path_argument("foo", "123")
        args = parse_args(self.env, [
            ("foo", int, Multiplicity.REQUIRED),
        ])
        assert_equal({"foo": 123}, args)

    def test_parse_invalid_int_arg(self) -> None:
        self.add_path_argument("foo", "bar")
        try:
            parse_args(self.env, [
                ("foo", int, Multiplicity.REQUIRED),
            ])
        except ArgumentsError as exc:
            assert_equal({
                "foo": "invalid literal for int() with base 10: 'bar'",
            }, exc.arguments)
        else:
            fail("ArgumentsError not raised")

    def test_required_argument_missing(self) -> None:
        try:
            parse_args(self.env, [
                ("foo", str, Multiplicity.REQUIRED),
            ])
        except ArgumentsError as exc:
            assert_equal({"foo": "mandatory argument missing"}, exc.arguments)
        else:
            fail("ArgumentsError not raised")

    def test_optional_argument(self) -> None:
        self.add_path_argument("foo", "bar")
        args = parse_args(self.env, [
            ("foo", str, Multiplicity.OPTIONAL),
        ])
        assert_equal({"foo": "bar"}, args)

    def test_optional_argument_missing(self) -> None:
        args = parse_args(self.env, [
            ("foo", str, Multiplicity.OPTIONAL),
        ])
        assert_equal({}, args)

    def test_any_argument_empty(self) -> None:
        args = parse_args(self.env, [
            ("foo", int, Multiplicity.ANY),
        ])
        assert_equal({"foo": []}, args)

    def test_any_argument(self) -> None:
        self.add_path_argument("foo", "123")
        self.add_path_argument("foo", "456")
        self.add_path_argument("foo", "789")
        args = parse_args(self.env, [
            ("foo", int, Multiplicity.ANY),
        ])
        assert_equal({"foo": [123, 456, 789]}, args)

    def test_required_any_argument(self) -> None:
        self.add_path_argument("foo", "123")
        self.add_path_argument("foo", "456")
        self.add_path_argument("foo", "789")
        args = parse_args(self.env, [
            ("foo", int, Multiplicity.REQUIRED_ANY),
        ])
        assert_equal({"foo": [123, 456, 789]}, args)

    def test_required_any_argument_missing(self) -> None:
        try:
            parse_args(self.env, [
                ("foo", int, Multiplicity.REQUIRED_ANY),
            ])
        except ArgumentsError as exc:
            assert_equal({"foo": "mandatory argument missing"}, exc.arguments)
        else:
            fail("ArgumentsError not raised")

    def test_urlencoded_post_request(self) -> None:
        self.env["REQUEST_METHOD"] = "POST"
        self.setup_urlencoded_request("foo", "bar")
        args = parse_args(self.env, [
            ("foo", str, Multiplicity.OPTIONAL),
        ])
        assert_equal({"foo": "bar"}, args)

    def test_urlencoded_post_request_with_umlauts(self) -> None:
        self.env["REQUEST_METHOD"] = "POST"
        self.setup_urlencoded_request("föo", "bär")
        args = parse_args(self.env, [
            ("föo", str, Multiplicity.OPTIONAL),
        ])
        assert_equal({"föo": "bär"}, args)

    def test_urlencoded_put_request(self) -> None:
        self.env["REQUEST_METHOD"] = "PUT"
        self.setup_urlencoded_request("foo", "bar")
        args = parse_args(self.env, [
            ("foo", str, Multiplicity.OPTIONAL),
        ])
        assert_equal({"foo": "bar"}, args)

    def test_urlencoded_put_request_with_umlauts(self) -> None:
        self.env["REQUEST_METHOD"] = "PUT"
        self.setup_urlencoded_request("föo", "bär")
        args = parse_args(self.env, [
            ("föo", str, Multiplicity.OPTIONAL),
        ])
        assert_equal({"föo": "bär"}, args)

    def test_multipart_post_request(self) -> None:
        self.env["REQUEST_METHOD"] = "POST"
        self.setup_multipart_request("foo", "bar")
        args = parse_args(self.env, [
            ("foo", str, Multiplicity.OPTIONAL),
        ])
        assert_equal({"foo": "bar"}, args)

    def test_multipart_post_request_with_umlauts(self) -> None:
        self.env["REQUEST_METHOD"] = "POST"
        self.setup_multipart_request("föo", "bär")
        args = parse_args(self.env, [
            ("föo", str, Multiplicity.OPTIONAL),
        ])
        assert_equal({"föo": "bär"}, args)

    def test_multipart_put_request(self) -> None:
        self.env["REQUEST_METHOD"] = "PUT"
        self.setup_multipart_request("foo", "bar")
        args = parse_args(self.env, [
            ("foo", str, Multiplicity.OPTIONAL),
        ])
        assert_equal({"foo": "bar"}, args)

    def test_multipart_put_request_with_umlauts(self) -> None:
        self.env["REQUEST_METHOD"] = "PUT"
        self.setup_multipart_request("föo", "bär")
        args = parse_args(self.env, [
            ("föo", str, Multiplicity.OPTIONAL),
        ])
        assert_equal({"föo": "bär"}, args)

    def test_multipart_post_request_with_file(self) -> None:
        self.env["REQUEST_METHOD"] = "POST"
        self.setup_multipart_file_request(
            "my-file", "my-file.txt", "content", "text/plain")
        args = parse_args(self.env, [
            ("my-file", "file", Multiplicity.REQUIRED),
        ])
        assert_in("my-file", args)
        f = args["my-file"]
        assert_equal("my-file.txt", f.filename)
        assert_equal("text/plain", f.content_type)
        assert_equal(b"content", f.read())

    def test_multipart_post_request_with_file_and_umlauts(self) -> None:
        self.env["REQUEST_METHOD"] = "POST"
        self.setup_multipart_file_request(
            "föo", "my-filé.txt", "cöntent", "text/plain; charset=utf-8")
        args = parse_args(self.env, [
            ("föo", "file", Multiplicity.REQUIRED),
        ])
        assert_in("föo", args)
        f = args["föo"]
        assert_equal("my-filé.txt", f.filename)
        assert_equal("text/plain", f.content_type)
        assert_equal("cöntent".encode("utf-8"), f.read())

    def test_multipart_put_request_with_file(self) -> None:
        self.env["REQUEST_METHOD"] = "PUT"
        self.setup_multipart_file_request(
            "my-file", "my-file.txt", "content", "text/plain")
        args = parse_args(self.env, [
            ("my-file", "file", Multiplicity.REQUIRED),
        ])
        assert_in("my-file", args)
        f = args["my-file"]
        assert_equal("my-file.txt", f.filename)
        assert_equal("text/plain", f.content_type)
        assert_equal(b"content", f.read())

    def test_read_file_as_value(self) -> None:
        self.env["REQUEST_METHOD"] = "POST"
        self.setup_multipart_file_request(
            "foo", "my-file.txt", "123", "text/plain")
        args = parse_args(self.env, [
            ("foo", int, Multiplicity.REQUIRED),
        ])
        assert_equal(123, args["foo"])

    def test_read_value_as_file(self) -> None:
        self.env["REQUEST_METHOD"] = "GET"
        self.add_path_argument("foo", "bar")
        args = parse_args(self.env, [
            ("foo", "file", Multiplicity.REQUIRED),
        ])
        assert_in("foo", args)
        f = args["foo"]
        assert_equal("", f.filename)
        assert_equal("application/octet-stream", f.content_type)
        assert_equal(b"bar", f.read())

    def test_read_value_as_file_with_umlauts(self) -> None:
        self.env["REQUEST_METHOD"] = "GET"
        self.add_path_argument("foo", "bär")
        args = parse_args(self.env, [
            ("foo", "file", Multiplicity.REQUIRED),
        ])
        assert_in("foo", args)
        f = args["foo"]
        assert_equal("", f.filename)
        assert_equal("application/octet-stream", f.content_type)
        assert_equal("bär".encode("utf-8"), f.read())



from http import HTTPStatus
from typing import Type
from unittest import TestCase

from asserts import assert_is, assert_equal
from werkzeug.wrappers import Request

from rouver.handler import RouteHandlerBase

from rouver_test.util import default_environment, StartResponse


class RouteHandlerBaseTest(TestCase):

    def setUp(self) -> None:
        self.request = Request(default_environment())
        self.start_response = StartResponse()

    def call_handler(self, handler_class: Type[RouteHandlerBase]) -> bytes:
        handler = handler_class(self.request, [], self.start_response)
        return b"".join(handler)

    def test_attributes(self) -> None:
        handler = RouteHandlerBase(self.request, ["foo"], self.start_response)
        assert_is(self.request, handler.request)
        assert_equal(["foo"], handler.path_args)
        assert_is(self.start_response, handler.start_response)

    def test_respond(self) -> None:
        class TestingHandler(RouteHandlerBase):
            def __iter__(self):
                return self.respond()
        response = self.call_handler(TestingHandler)
        self.start_response.assert_status(HTTPStatus.OK)
        assert_equal(b"", response)



#!/usr/bin/python

import os

from setuptools import setup


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="rouver",
    version="0.2.1",
    description="A microframework",
    long_description=read("README.rst"),
    author="Sebastian Rittau",
    author_email="srittau@rittau.biz",
    url="https://github.com/srittau/rouver",
    packages=["rouver", "rouver_test"],
    install_requires=["werkzeug >= 0.12.0"],
    tests_require=["asserts >= 0.7.0, < 0.8"],
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Internet :: WWW/HTTP :: WSGI",
    ]
)
